#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/ethernet.h>
#include <netinet/in.h>

#include "asm_val.h"
#include "lib.h"

struct pkt_asm_val {
  pkt_asm_val_type_t type;
  union {
    int integer;
    struct ether_addr macaddr;
    struct in_addr ipaddr;
    struct in6_addr ipv6addr;
    char *string;
  } value;
};

static pkt_asm_val_t pkt_asm_val_create(pkt_asm_val_type_t type)
{
  pkt_asm_val_t val;

  val = malloc(sizeof(*val));
  if (val == NULL)
    error_exit("Out of memory.\n");

  memset(val, 0, sizeof(*val));

  val->type = type;

  return val;
}

pkt_asm_val_t pkt_asm_val_create_integer(int integer)
{
  pkt_asm_val_t val;
  val = pkt_asm_val_create(PKT_ASM_VAL_TYPE_INTEGER);
  pkt_asm_val_set_value_integer(val, integer);
  return val;
}

pkt_asm_val_t pkt_asm_val_create_macaddr(struct ether_addr *macaddr)
{
  pkt_asm_val_t val;
  val = pkt_asm_val_create(PKT_ASM_VAL_TYPE_MACADDR);
  pkt_asm_val_set_value_macaddr(val, macaddr);
  return val;
}

pkt_asm_val_t pkt_asm_val_create_ipaddr(struct in_addr *ipaddr)
{
  pkt_asm_val_t val;
  val = pkt_asm_val_create(PKT_ASM_VAL_TYPE_IPADDR);
  pkt_asm_val_set_value_ipaddr(val, ipaddr);
  return val;
}

pkt_asm_val_t pkt_asm_val_create_ipv6addr(struct in6_addr *ipv6addr)
{
  pkt_asm_val_t val;
  val = pkt_asm_val_create(PKT_ASM_VAL_TYPE_IPV6ADDR);
  pkt_asm_val_set_value_ipv6addr(val, ipv6addr);
  return val;
}

pkt_asm_val_t pkt_asm_val_create_string(char *string)
{
  pkt_asm_val_t val;
  val = pkt_asm_val_create(PKT_ASM_VAL_TYPE_STRING);
  val->value.string = NULL;
  pkt_asm_val_set_value_string(val, string);
  return val;
}

pkt_asm_val_t pkt_asm_val_destroy(pkt_asm_val_t val)
{
  if (val) {
    switch (val->type) {
    case PKT_ASM_VAL_TYPE_STRING:
      if (val->value.string) free(val->value.string);
      break;
    default:
      break;
    }
    free(val);
  }
  return NULL;
}

pkt_asm_val_type_t pkt_asm_val_get_type(pkt_asm_val_t val)
{
  return val->type;
}

int pkt_asm_val_get_value_integer(pkt_asm_val_t val)
{
  if (val->type != PKT_ASM_VAL_TYPE_INTEGER)
    return -1;
  return val->value.integer;
}

struct ether_addr *pkt_asm_val_get_value_macaddr(pkt_asm_val_t val)
{
  if (val->type != PKT_ASM_VAL_TYPE_MACADDR)
    return NULL;
  return &val->value.macaddr;
}

struct in_addr *pkt_asm_val_get_value_ipaddr(pkt_asm_val_t val)
{
  if (val->type != PKT_ASM_VAL_TYPE_IPADDR)
    return NULL;
  return &val->value.ipaddr;
}

struct in6_addr *pkt_asm_val_get_value_ipv6addr(pkt_asm_val_t val)
{
  if (val->type != PKT_ASM_VAL_TYPE_IPV6ADDR)
    return NULL;
  return &val->value.ipv6addr;
}

char *pkt_asm_val_get_value_string(pkt_asm_val_t val)
{
  if (val->type != PKT_ASM_VAL_TYPE_STRING)
    return NULL;
  return val->value.string;
}

int pkt_asm_val_set_value_integer(pkt_asm_val_t val, int integer)
{
  if (val->type != PKT_ASM_VAL_TYPE_INTEGER)
    return -1;
  val->value.integer = integer;
  return 0;
}

int pkt_asm_val_set_value_macaddr(pkt_asm_val_t val, struct ether_addr *macaddr)
{
  if (val->type != PKT_ASM_VAL_TYPE_MACADDR)
    return -1;
  memcpy(&val->value.macaddr, macaddr, ETHER_ADDR_LEN);
  return 0;
}

int pkt_asm_val_set_value_ipaddr(pkt_asm_val_t val, struct in_addr *ipaddr)
{
  if (val->type != PKT_ASM_VAL_TYPE_IPADDR)
    return -1;
  memcpy(&val->value.ipaddr, ipaddr, sizeof(*ipaddr));
  return 0;
}

int pkt_asm_val_set_value_ipv6addr(pkt_asm_val_t val, struct in6_addr *ipv6addr)
{
  if (val->type != PKT_ASM_VAL_TYPE_IPV6ADDR)
    return -1;
  memcpy(&val->value.ipv6addr, ipv6addr, sizeof(*ipv6addr));
  return 0;
}

int pkt_asm_val_set_value_string(pkt_asm_val_t val, char *string)
{
  if (val->type != PKT_ASM_VAL_TYPE_STRING)
    return -1;
  if (val->value.string)
    free(val->value.string);
  val->value.string = strdup(string);
  if (val->value.string == NULL)
    error_exit("Out of memory.\n");
  return 0;
}
