#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include "filter.h"
#include "lib.h"

#define REJECT 0
#define PASS   1

/* Select filter rule (choose PASS or REJECT) */
#define FILTER_RULE_MATCH   REJECT
#define FILTER_RULE_UNMATCH PASS

#define   MATCH(r) do { (r) |= 1; } while (0)
#define UNMATCH(r) do { /* nop */ } while (0)

static int filter_arp(int r, char *buffer, int size)
{
  struct arphdr arphdr;
  char *p, *smac, *tmac, *sip, *tip;
  struct arpdata {
    union {
      char *octet;
      struct ether_addr *addr;
    } sender_macaddr;
    union {
      char *octet;
      struct ether_addr *addr;
    } target_macaddr;
    struct in_addr sender_ipaddr;
    struct in_addr target_ipaddr;
  } arpdata;

  if (size < sizeof(arphdr))
    return -1;
  memcpy(&arphdr, buffer, sizeof(arphdr));

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  switch (ntohs(arphdr.ar_op)) { /* operation */
  case ARPOP_REQUEST: UNMATCH(r); break;
  case ARPOP_REPLY:   UNMATCH(r); break;
  default:            UNMATCH(r); break;
  }
  switch (ntohs(arphdr.ar_hrd)) { /* hardware */
  case ARPHRD_ETHER: UNMATCH(r); break;
  default:           UNMATCH(r); break;
  }
  switch (ntohs(arphdr.ar_pro)) { /* protocol */
  case ETHERTYPE_IP: UNMATCH(r); break;
  default:           UNMATCH(r); break;
  }
  switch (arphdr.ar_hln) { /* hard size */
  case ETHER_ADDR_LEN: UNMATCH(r); break;
  default:             UNMATCH(r); break;
  }
  switch (arphdr.ar_pln) { /* proto size */
  case sizeof(in_addr_t): UNMATCH(r); break;
  default:                UNMATCH(r); break;
  }

  if (size < sizeof(struct arphdr) + arphdr.ar_hln * 2 + arphdr.ar_pln * 2)
    return -1;

  p = buffer;
  p += sizeof(struct arphdr); smac = p;
  p += arphdr.ar_hln; sip  = p;
  p += arphdr.ar_pln; tmac = p;
  p += arphdr.ar_hln; tip  = p;

  arpdata.sender_macaddr.octet = smac;
  arpdata.target_macaddr.octet = tmac;
  memcpy(&arpdata.sender_ipaddr.s_addr, sip, sizeof(in_addr_t));
  memcpy(&arpdata.target_ipaddr.s_addr, tip, sizeof(in_addr_t));

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  if (!memcmp(arpdata.sender_macaddr.octet, /* sender MAC */
	      "\xFF\xFF\xFF\xFF\xFF\xFF", ETHER_ADDR_LEN))
    UNMATCH(r);
  if (!memcmp(arpdata.target_macaddr.octet, /* target MAC */
	      "\xFF\xFF\xFF\xFF\xFF\xFF", ETHER_ADDR_LEN))
    UNMATCH(r);

  if (ntohl(arpdata.sender_ipaddr.s_addr) == 0xC0A80101) /* sender IP */
    UNMATCH(r);
  if (ntohl(arpdata.target_ipaddr.s_addr) == 0xC0A80102) /* target IP */
    UNMATCH(r);

  return r;
}

static int filter_icmp(int r, char *buffer, int size, int total_size)
{
  struct icmp *icmphdr;

  if ((size < total_size) || (total_size < sizeof(*icmphdr)))
    return -1;
  icmphdr = (struct icmp *)buffer;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  switch (icmphdr->icmp_type) { /* type */
  case ICMP_ECHO:      UNMATCH(r); break;
  case ICMP_ECHOREPLY: UNMATCH(r); break;
  default:             UNMATCH(r); break;
  }
  switch (icmphdr->icmp_code) { /* code */
  default:             UNMATCH(r); break;
  }
  if (ntohs(icmphdr->icmp_cksum) == 0) UNMATCH(r); /* checksum */

  return r;
}

static int filter_igmp(int r, char *buffer, int size, int total_size)
{
  struct igmp *igmphdr;

  if ((size < total_size) || (total_size < sizeof(*igmphdr)))
    return -1;
  if (total_size < IGMP_MINLEN)
    return -1;
  igmphdr = (struct igmp *)buffer;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  switch (igmphdr->igmp_type) { /* type */
  default: UNMATCH(r); break;
  }
  switch (igmphdr->igmp_code) { /* code */
  default: UNMATCH(r); break;
  }
  if (ntohs(igmphdr->igmp_cksum) == 0) UNMATCH(r); /* checksum */
  if (ntohl(igmphdr->igmp_group.s_addr) == 0xE0000001) /* group (224.0.0.1) */
    UNMATCH(r);
  if (ntohl(igmphdr->igmp_group.s_addr) == 0xEFFFFFFF) /* (239.255.255.255) */
    UNMATCH(r);

  return r;
}

static int filter_tcp(int r, char *buffer, int size, int total_size)
{
  struct tcphdr *tcphdr;

  if ((size < total_size) || (total_size < sizeof(*tcphdr)))
    return -1;
  tcphdr = (struct tcphdr *)buffer;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  if (ntohs(tcphdr->th_sport) == 21) UNMATCH(r); /* src port (FTP) */
  if (ntohs(tcphdr->th_dport) == 80) UNMATCH(r); /* dst port (HTTP) */
  if (ntohl(tcphdr->th_seq  ) ==  0) UNMATCH(r); /* seq number */
  if (ntohl(tcphdr->th_ack  ) ==  0) UNMATCH(r); /* ack number */
  if ((tcphdr->th_off << 2)   ==  0) UNMATCH(r); /* offset */
  if (tcphdr->th_flags) { /* flags */
    if (tcphdr->th_flags & TH_FIN ) UNMATCH(r); /* FIN */
    if (tcphdr->th_flags & TH_SYN ) UNMATCH(r); /* SYN */
    if (tcphdr->th_flags & TH_RST ) UNMATCH(r); /* RST */
    if (tcphdr->th_flags & TH_PUSH) UNMATCH(r); /* PSH */
    if (tcphdr->th_flags & TH_ACK ) UNMATCH(r); /* ACK */
    if (tcphdr->th_flags & TH_URG ) UNMATCH(r); /* URG */
  }
  if (ntohs(tcphdr->th_win  ) ==  0) UNMATCH(r); /* window */
  if (ntohs(tcphdr->th_sum  ) ==  0) UNMATCH(r); /* checksum */

  return r;
}

static int filter_udp(int r, char *buffer, int size, int total_size)
{
  struct udphdr *udphdr;

  if ((size < total_size) || (total_size < sizeof(*udphdr)))
    return -1;
  udphdr = (struct udphdr *)buffer;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  if (ntohs(udphdr->uh_sport) == 53) UNMATCH(r); /* src port (DNS) */
  if (ntohs(udphdr->uh_dport) == 67) UNMATCH(r); /* dst port (DHCP) */
  if (ntohs(udphdr->uh_ulen ) ==  0) UNMATCH(r); /* length */
  if (ntohs(udphdr->uh_sum  ) ==  0) UNMATCH(r); /* checksum */

  return r;
}

static int filter_ip(int r, char *buffer, int size)
{
  static char *pktbuf = NULL;
  static int bufsize = 0;
  char *p;
  int s;
  struct ip *iphdr;
  int hdrsize, paysize;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (hdrsize < sizeof(*iphdr))
    return -1;
  if (paysize < 0)
    return -1;
  if (size < hdrsize + paysize)
    return -1;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  if (ntohs(iphdr->ip_len) == 0) UNMATCH(r); /* total size */
  if (ntohs(iphdr->ip_id ) == 0) UNMATCH(r); /* ID */
  if (ntohs(iphdr->ip_off) == 0) UNMATCH(r); /* fragment */
  if (      iphdr->ip_ttl  == 0) UNMATCH(r); /* TTL */
  if (ntohs(iphdr->ip_sum) == 0) UNMATCH(r); /* checksum */

  if (ntohl(iphdr->ip_src.s_addr) == 0xC0A80101) UNMATCH(r); /* src IP addr */
  if (ntohl(iphdr->ip_dst.s_addr) == 0xC0A80102) UNMATCH(r); /* dst IP addr */

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP: r = filter_icmp(r, p, s, paysize); break;
  case IPPROTO_IGMP: r = filter_igmp(r, p, s, paysize); break;
  case IPPROTO_TCP:  r = filter_tcp( r, p, s, paysize); break;
  case IPPROTO_UDP:  r = filter_udp( r, p, s, paysize); break;
  default: UNMATCH(r); break;
  }

  if (r < 0)
    return r;

  return r;
}

static int filter_icmp6(int r, char *buffer, int size, int total_size)
{
  struct icmp6_hdr *icmp6hdr;

  if ((size < total_size) || (total_size < sizeof(*icmp6hdr)))
    return -1;
  icmp6hdr = (struct icmp6_hdr *)buffer;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  switch (icmp6hdr->icmp6_type) { /* type */
  case ICMP6_ECHO_REQUEST: UNMATCH(r); break;
  case ICMP6_ECHO_REPLY:   UNMATCH(r); break;
  default:                 UNMATCH(r); break;
  }
  switch (icmp6hdr->icmp6_code) { /* code */
  default:                 UNMATCH(r); break;
  }
  if (ntohs(icmp6hdr->icmp6_cksum) == 0) UNMATCH(r); /* checksum */

  return r;
}

static int filter_ip6(int r, char *buffer, int size)
{
  static char *pktbuf = NULL;
  static int bufsize = 0;
  char *p;
  int s;
  struct ip6_hdr *ip6hdr;
  int hdrsize, paysize;
  int nexthdr, exthdrsize;
  struct ip6_ext *ip6exthdr;
  struct ip6_rthdr *ip6rthdr;
  struct ip6_frag *ip6fraghdr;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*ip6hdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  ip6hdr = (struct ip6_hdr *)pktbuf;

  hdrsize = sizeof(struct ip6_hdr);
  paysize = ntohs(ip6hdr->ip6_plen);

  if (size < hdrsize + paysize)
    return -1;

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  if (ip6hdr->ip6_hlim == 0) UNMATCH(r); /* hop limit */

  if (!memcmp(&ip6hdr->ip6_src.s6_addr, /* src IPv6 addr */
	      "\xFE\xC0\0\0\0\0\0\0\0\0\0\0\0\0\0\x01",
	      sizeof(struct in6_addr))) UNMATCH(r);
  if (!memcmp(&ip6hdr->ip6_dst.s6_addr, /* dst IPv6 addr */
	      "\xFE\xC0\0\0\0\0\0\0\0\0\0\0\0\0\0\x02",
	      sizeof(struct in6_addr))) UNMATCH(r);

  p = pktbuf + hdrsize;
  s = size   - hdrsize;

  nexthdr = ip6hdr->ip6_nxt;

  while (1) {
    exthdrsize = 0;
    switch (nexthdr) {
    case IPPROTO_HOPOPTS:
    case IPPROTO_DSTOPTS:
      if (paysize < sizeof(*ip6exthdr))
	return -1;
      ip6exthdr = (struct ip6_ext *)p;
      nexthdr = ip6exthdr->ip6e_nxt;
      exthdrsize = (ip6exthdr->ip6e_len + 1) << 3;
      break;
    case IPPROTO_ROUTING:
      if (paysize < sizeof(*ip6rthdr))
	return -1;
      ip6rthdr = (struct ip6_rthdr *)p;
      nexthdr = ip6rthdr->ip6r_nxt;
      exthdrsize = (ip6rthdr->ip6r_len + 1) << 3;
      break;
    case IPPROTO_FRAGMENT:
      if (paysize < sizeof(*ip6fraghdr))
	return -1;
      ip6fraghdr = (struct ip6_frag *)p;
      nexthdr = ip6fraghdr->ip6f_nxt;
      exthdrsize = sizeof(*ip6fraghdr);
      break;
    case IPPROTO_NONE:
    default:
      break;
    }
    if (exthdrsize == 0)
      break;
    if (paysize < exthdrsize)
      return -1;
    p       += exthdrsize;
    s       -= exthdrsize;
    paysize -= exthdrsize;
  }

  switch (nexthdr) {
  case IPPROTO_ICMPV6: r = filter_icmp6(r, p, s, paysize); break;
  case IPPROTO_TCP:    r = filter_tcp(  r, p, s, paysize); break;
  case IPPROTO_UDP:    r = filter_udp(  r, p, s, paysize); break;
  default: UNMATCH(r); break;
  }

  if (r < 0)
    return r;

  return r;
}

int pkt_filter_ethernet(char *buffer, int size)
{
  int r = 0;
  char *p;
  int s;
  struct ether_header ehdr;
  struct {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;

  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);

  /*********************************/
  /* SETUP FILTER RULE AS YOU LIKE */
  /*********************************/

  if (!memcmp(ehdr.ether_shost, "\xFF\xFF\xFF\xFF\xFF\xFF", /* src MAC addr */
	      ETHER_ADDR_LEN)) UNMATCH(r);
  if (!memcmp(ehdr.ether_dhost, "\xFF\xFF\xFF\xFF\xFF\xFF", /* dst MAC addr */
	      ETHER_ADDR_LEN)) UNMATCH(r);
  if (ntohs(ehdr.ether_type) == 0) /* type */
    UNMATCH(r);

  type = ntohs(ehdr.ether_type);
  p = buffer + ETHER_HDR_LEN;
  s = size   - ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif
#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6 0x86DD
#endif

  while (type == ETHERTYPE_VLAN) {
    if (s < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, p, sizeof(vlantag));
    if (ntohs(vlantag.tag) == 0) /* VLAN tag */
      UNMATCH(r);
    if (ntohs(vlantag.proto) == 0) /* VLAN protocol */
      UNMATCH(r);
    type = ntohs(vlantag.proto);
    p += sizeof(vlantag);
    s -= sizeof(vlantag);
  }

  switch (type) {
  case ETHERTYPE_ARP:  r = filter_arp(r, p, s); break;
  case ETHERTYPE_IP:   r = filter_ip( r, p, s); break;
  case ETHERTYPE_IPV6: r = filter_ip6(r, p, s); break;
  default: UNMATCH(r); break;
  }

  if (r < 0)
    return r;

  return (r == 0) ? FILTER_RULE_UNMATCH : FILTER_RULE_MATCH;
}
